/*
 * Copyright (c) 2022 Reza Mahdi. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are hereby granted under terms of BSD 3-Clause License.
 * See LICENSE for more information.
 */

#ifndef DNSPIDER_LIST_H
#define DNSPIDER_LIST_H

#include <memory.h>
#include <stddef.h>

/**
 * @brief head field for creating a doublly linked list.
 */
typedef struct list_head_s
{
  struct list_head_s *next; /**< pointer to next element */
  struct list_head_s *prev; /**< pointer to prev. element */
} list_head_t;

/**
 * @brief main struct of a linked list
 */
typedef struct
{
  list_head_t *head; /**< link to head of list */
  list_head_t *tail; /**< link to tail of list */
  size_t elements;   /**< number of elements in list */
} list_t;

/**
 * @brief initialize a linked list
 *
 * @param list the list to initialize
 */
void list_init (list_t *list);

/**
 * @brief append a new element to list
 *
 * @param list list to append the element
 * @param element the element to append
 */
void list_append (list_t *list, list_head_t *element);

/**
 * @brief prepend a new element to list
 *
 * @param list list to prepend the element
 * @param element the element to prepend
 */
void list_prepend (list_t *list, list_head_t *element);

/**
 * @brief pop item from back (head) of list
 *
 * @param list list to pop item from
 * @return list_head_t* item at back
 */
list_head_t *list_pop_back (list_t *list);

/**
 * @brief pop item from front (tail) of list
 *
 * @param list list to pop item from
 * @return list_head_t* item at front
 */
list_head_t *list_pop_front (list_t *list);

/**
 * @brief check if list is empty or not
 *
 * @param list the list to check
 * @return 1 if list is empty, 0 otherwise
 */
int list_is_empty (list_t *list);

/**
 * @brief number of elements in this list
 *
 * @param list
 * @return size_t
 */
size_t list_count (list_t *list);

/**
 * @brief implement a foreach loop for list
 *
 * @param list the list to iterate on
 * @param item name of iterating item
 * @param type type of items in list
 * @param field name of list_head_t defined in type
 */
#define list_foreach(list, item, type, field)                                 \
  for (type *item = (type *)list->head; item->field.next != NULL;             \
       item = (type *)item->field.next)
#endif