/*
 * Copyright (c) 2022 Reza Mahdi. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are hereby granted under terms of BSD 3-Clause License.
 * See LICENSE for more information.
 */

#include <errno.h>
#include <getopt.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/sysinfo.h>

#include "list.h"
#include "parser.h"
#include "queue.h"
#include "worker.h"

#if defined(HAVE_CONFIG_H)
#include <config.h>
#endif

static const struct option long_options[]
    = { { "input", required_argument, NULL, 'i' },
        { "servers", required_argument, NULL, 's' },
        { "workers", required_argument, NULL, 'w' },
        { "loop", required_argument, NULL, 'l' },
        { "out", required_argument, NULL, 'o' },
        { "outform", required_argument, NULL, 'f' },
        { "benchmark", optional_argument, NULL, 'b' },
        { "verbose", no_argument, NULL, 'v' },
        { "help", no_argument, NULL, 'h' },
        { NULL, 0, NULL, 0 } };

static const char *short_options = "t:d:s:w:l:p:o:f:b:h";

void
print_help ()
{
  printf (
      "dnspider - DNS subdomain scanner\n"
      "\n  Usage: dnspider [options] <target-selection>\n"
      "  Available options:\n"
      "\t-i, --input         Add dictionary file to sources\n"
      "\t-s, --servers       Comma-separated list of DNS servers\n"
      "\t-w, --workers       Number of worker threads to use (default: %d)\n"
      "\t-l, --loop          Select loop backend. Separate loop options "
      "with comma\n"
      "\t-o, --out           Output file name\n"
      "\t-f, --outform       Out file format\n"
      "\t-b, --benchmark     Measure timings to do a benchmark beside scan\n"
      "\t-v, --verbose       Increase logging verbosity\n"
      "\t-h, --help          Print this help message\n",
      get_nprocs ());
  printf ("Please report any bug to '%s'\n", PACKAGE_BUGREPORT);
}

/* main workers */
static worker_pool_t workers_pool;

/**
 * @brief initialize workers
 *
 * @param num_workers number of workers
 * @return int
 */
static int init_workers (uint32_t num_workers);

/**
 * @brief Open input files
 *
 * @return int
 */
static int open_inputs ();

/**
 * @brief dispatch subdomains to workers
 *
 * read subdomains form input and dispatch them to queue
 *
 * @return int
 */
static int dispatch ();

int
main (int argc, char *const *argv)
{
  int arg;
  char *target_name;
  char *subdomains_file;
  int workers = get_nprocs ();
  int verbosity = 0;
  char *output;
  char *outform;
  char *selector = NULL;

  while ((arg = getopt_long (argc, argv, short_options, long_options, NULL))
         != -1)
    {
      switch (arg)
        {
        case 't':
          target_name = optarg;
          break;
        case 'd':
          subdomains_file = optarg;
          break;
        case 's':
          break;
        case 'w':
          workers = atoi (optarg);
          break;
        case 'o':
          output = optarg;
          break;
        case 'h':
          print_help ();
          exit (0);
          break;
        case 'v':
          verbosity++;
          break;
        }
    }

  if (optind < argc)
    {
      printf ("selector string: %s\n", argv[optind]);
      selector = argv[optind];
    }
  else
    {
      printf ("No selector specified. exit.\n");
      exit (1);
    }
  
  if(parser_init()) {
    printf("could not init parser\n");
  }
  list_t lexems;
  parse_target(selector, &lexems);

  if (init_workers (workers))
    {
      printf ("Initializing workers: %s\n", strerror (errno));
    }
}

static int
init_workers (uint32_t workers)
{
  if (queue_spsc_init (&workers_pool.job_queue, 1024, sizeof (void *)))
    {
      return -1;
    }
  /* TODO: initialize workers */
  list_init (&workers_pool.workers);
  return 0;
}