/*
 * Copyright (c) 2022 Reza Mahdi. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are hereby granted under terms of BSD 3-Clause License.
 * See LICENSE for more information.
 */

#include "parser.h"
#include "list.h"
#include <errno.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(HAVE_CONFIG_H)
#include <config.h>
#endif

#if defined(HAVE_PCRE2)
#define PCRE2_CODE_UNIT_WIDTH 8
#include <pcre2.h>

static pcre2_code *regex;
static pcre2_code *range_regex;
static pcre2_code *add_line_regex;
static pcre2_code *sub_line_regex;
static pcre2_code *add_label_regex;

#define REGEX_COMPILE_AND_CHECK(regex_, pattern_)                             \
  (regex_ = pcre2_compile ((PCRE2_SPTR)pattern_, PCRE2_ZERO_TERMINATED, 0,    \
                           &error, &offset, 0))                               \
      == NULL

#define FULL_MATCH(f_regex, str, len)                                         \
  int rc;                                                                     \
  PCRE2_SIZE *ovector, substring_length;                                      \
  pcre2_match_data *match_data;                                               \
  PCRE2_SPTR substring_start;                                                 \
  match_data = pcre2_match_data_create_from_pattern (f_regex, NULL);          \
  rc = pcre2_match (f_regex, (PCRE2_SPTR)str, len, 0, 0, match_data, NULL);   \
  if (rc < 0)                                                                 \
    {                                                                         \
      switch (rc)                                                             \
        {                                                                     \
        case PCRE2_ERROR_NOMATCH:                                             \
          return 0;                                                           \
          break;                                                              \
        default:                                                              \
          return -1;                                                          \
          break;                                                              \
        }                                                                     \
    }                                                                         \
  else                                                                        \
    {                                                                         \
      return 1;                                                               \
    }
#else

#include <regex.h>

static regex_t regex;
static regex_t range_regex;
static regex_t add_line_regex;
static regex_t sub_line_regex;
static regex_t add_label_regex;

#define REGEX_COMPILE_AND_CHECK(regex_, pattern_)                             \
  do                                                                          \
    {                                                                         \
      if (regcomp (&regex, CAPTURE_REGEX, 0) == 0)                            \
        {                                                                     \
          return 0;                                                           \
        }                                                                     \
      else                                                                    \
        {                                                                     \
          return 1;                                                           \
        }                                                                     \
    }                                                                         \
  while (0)
#endif

#define CAPTURE_REGEX "\\[[^]]*\\]"
#define RANGE_REGEX "([0-9]+:)?[0-9]*~[0-9]*"
#define ADD_LINE_REGEX "\\+[0-9]+"
#define SUB_LINE_REGEX "\\-[0-9]+"
#define ADD_LABEL_REGEX "[A-Za-z]+"

/* check current lexem to be a range selector */
static int is_range_select (const char *buffer, size_t len);
/* check current lexem to be a line addition */
static int is_add_line (const char *buffer, size_t len);
/* check current lexem to be a line substraction */
static int is_sub_line (const char *buffer, size_t len);
/* check current lexem to be a immediate lable */
static int is_imm_lable (const char *buffer, size_t len);
/* get type of current lexem */
static int lexem_type (const char *buffer, size_t len);
/* get next lexem */
static int next_lexem (const char *buffer, size_t remaining);
/* convert ascii to int by len */
static int atoin (const char *str, int len);

int
parser_init ()
{
  int error;
  size_t offset;

  if (REGEX_COMPILE_AND_CHECK (regex, CAPTURE_REGEX))
    {
    }
  if (REGEX_COMPILE_AND_CHECK (range_regex, RANGE_REGEX))
    {
    }
  if (REGEX_COMPILE_AND_CHECK (add_line_regex, ADD_LINE_REGEX))
    {
    }
  if (REGEX_COMPILE_AND_CHECK (sub_line_regex, SUB_LINE_REGEX))
    {
    }
  if (REGEX_COMPILE_AND_CHECK (add_label_regex, ADD_LABEL_REGEX))
    {
    }
}

int
parse_target (const char *target, list_t *lexems)
{
  list_init (lexems);
#if defined(HAVE_PCRE2)
  int rc;
  PCRE2_SIZE *ovector, substring_length;
  pcre2_match_data *match_data;
  PCRE2_SPTR substring_start;
  size_t lexem_len, remaining_len;
  unsigned int ltype = PARSER_UNKWON;
  char *lexem_start;
  parser_node_t *node;

  match_data = pcre2_match_data_create_from_pattern (regex, NULL);
  rc = pcre2_match (regex, (PCRE2_SPTR)target, PCRE2_ZERO_TERMINATED, 0, 0,
                    match_data, NULL);

  if (rc < 0)
    {
      switch (rc)
        {
        case PCRE2_ERROR_NOMATCH:
          break;
        default:
          return -1;
        }
      pcre2_match_data_free (match_data);
    }

  ovector = pcre2_get_ovector_pointer (match_data);
  for (int i = 0; i < rc; i++)
    {
      substring_start = target + ovector[2 * i];
      substring_length = ovector[2 * i + 1] - ovector[2 * i];
      /* TODO: parser options */
      if ((const char *)substring_start > target)
        {
          // current substring has a head (may be a const label)
        }

      // iterate over lexems
      lexem_start = (char *)substring_start;
      lexem_len = 0;
      remaining_len = substring_length;
      while ((lexem_len = next_lexem (lexem_start, remaining_len)) > 0)
        {
          lexem_start++; // skip '[' and ',' ...
          lexem_len--;   // ... and decrease len by one
          node = malloc (sizeof (parser_node_t));
          ltype = lexem_type (lexem_start, lexem_len);
          switch (ltype)
            {
            case PARSER_RANGE:
              node->type = PARSER_RANGE;

              /* Location of ':' and '~' */
              char *col_pos = memchr (lexem_start, ':', lexem_len);
              char *til_pos = memchr (lexem_start, '~', lexem_len);

              /* file index is at start and before location of ':' */
              node->range.file_index
                  = atoin (lexem_start, col_pos - lexem_start);

              /* start line is zero by default */
              node->range.start_line = 0;
              node->range.start_line
                  += atoin (col_pos + 1, til_pos - col_pos - 1);

              node->range.end_line = 0;
              if (til_pos == lexem_start + lexem_len)
                node->range.end_line = -1;
              else
                node->range.end_line
                    = atoin (til_pos + 1, lexem_start + lexem_len - til_pos);
              list_append (lexems, &node->head);
              break;
            case PARSER_ADD_LINE:
              node->type = PARSER_ADD_LINE;
              node->add_line.file = 0;
              node->add_line.line = atoin (lexem_start + 1, lexem_len - 1);
              list_append (lexems, &node->head);
              break;
            case PARSER_SUB_LINE:
              node->type = PARSER_SUB_LINE;
              node->sub_line.file = 0;
              node->sub_line.line = atoin (lexem_start + 1, lexem_len - 1);
              list_append (lexems, &node->head);
              break;
            case PARSER_ADD_LABEL:
              node->type = PARSER_ADD_LABEL;
              node->add_lable.lable = lexem_start;
              node->add_lable.len = lexem_len;
              list_append (lexems, &node->head);
              break;
            case PARSER_UNKWON:
              break;
            }
          // move to next lexem
          lexem_start += lexem_len;
          remaining_len -= lexem_len;
        }
    }
  pcre2_match_data_free (match_data);

  return 0;
#else
  /*regexec(&regex, target, size_t __nmatch, regmatch_t *restrict __pmatch,
          int __eflags);*/
#endif
}

/******* Lexing *************************************************************/
static int
is_range_select (const char *buffer, size_t len)
{
  FULL_MATCH (range_regex, buffer, len);
}

static int
is_add_line (const char *buffer, size_t len)
{
  FULL_MATCH (add_line_regex, buffer, len)
}

static int
is_sub_line (const char *buffer, size_t len)
{
  FULL_MATCH (sub_line_regex, buffer, len)
}

static int
is_imm_lable (const char *buffer, size_t len)
{
  FULL_MATCH (add_label_regex, buffer, len)
}

static int
lexem_type (const char *buffer, size_t len)
{
  if (is_range_select (buffer, len))
    return PARSER_RANGE;
  if (is_add_line (buffer, len))
    return PARSER_ADD_LINE;
  if (is_sub_line (buffer, len))
    return PARSER_SUB_LINE;
  if (is_imm_lable (buffer, len))
    return PARSER_ADD_LABEL;
  return PARSER_UNKWON;
}

static int
next_lexem (const char *buffer, size_t remaining)
{
  size_t current_char = 0;

  if (buffer[current_char] == ',')
    current_char++;
  // move forward to reach ',' or ']' (end of selector string)
  while (current_char <= remaining)
    {
      if (buffer[current_char] == ',' || buffer[current_char] == ']')
        return current_char;
      current_char++;
    }
  return -1;
}

static int
atoin (const char *str, int len)
{
  int i;
  int ret = 0;
  for (i = 0; i < len; ++i)
    {
      ret = ret * 10 + (str[i] - '0');
    }
  return ret;
}
