/*
 * Copyright (c) 2022 Reza Mahdi. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are hereby granted under terms of BSD 3-Clause License.
 * See LICENSE for more information.
 */

/**
 * @file parser.h
 * parsing facilities for parsing query making rule. target selecting rule on
 * command line is as follow:
 *
 *    '[' (RANGE-SELECTION | IMMEDIATE-ID | INCLUDE-SPEC | EXCLUDE-SPEC)+ ']'
 *
 * RANGE-SELECTION is combination of a sequencial numeric id assigned to input
 * dictionary file starting from 1 and a range selection that is a range of
 * lines in desired input file.
 * this range is in form:
 *
 *    RANGE-SELECTION = INPUT-FILE ':' [start-line] '~' [end-line]
 *
 * include-exclude make it possible to add or remove an specific line, whether
 * the line is in range or not. adding line is possible with adding a '+', and
 * removing is possible with prefixing with '-'. if plus or minuse sign is not
 * prefixed, this part is assumed a real 'lable'. this is usefull when you just
 * need to scan a short list of lables like '.org' and '.com' suffixes only.
 * this rule is applicable when there is a prefix but token after it is not
 * numerical.
 *
 * example:
 *            [1:0~30,-12,+63].example.[come,org,co,dev]
 *                ^    ^   ^                  ^
 *                |    |   |                  |
 *      select lines 0 to 30 from input 1    append these lables to seacrh
 *                     |   |
 *                  except line 12
 *                         |
 *                      and add line 63 to search
 */

#ifndef DNSPIDER_PARSER_H_
#define DNSPIDER_PARSER_H_

#include <stddef.h>
#if defined(HAVE_CONFIG_H)
#include <config.h>
#endif

#if defined(HAVE_PCRE2)
#define PCRE2_CODE_UNIT_WIDTH 8
#include <pcre2.h>
#else
#include <regex.h>
#endif

#include "list.h"

typedef struct
{
  unsigned file_index;
  unsigned start_line;
  unsigned end_line;
} parser_range_t;

typedef struct
{
  unsigned file;
  unsigned line;
} parser_add_line_t;

typedef parser_add_line_t parser_sub_line_t;

typedef struct
{
  const char *lable;
  size_t len;
} parser_add_lable_t;

typedef parser_add_lable_t parser_const_lable_t;

enum parser_segment_t
{
  PARSER_UNKWON,
  PARSER_RANGE,
  PARSER_ADD_LINE,
  PARSER_SUB_LINE,
  PARSER_ADD_LABEL
};

typedef struct
{
  list_head_t head;
  enum parser_segment_t type;
  union
  {
    parser_range_t range;
    parser_add_line_t add_line;
    parser_sub_line_t sub_line;
    parser_add_lable_t add_lable;
  };
} parser_node_t;

/**
 * @brief initialize parser.
 */
int parser_init ();

int parse_target (const char *target, list_t *lexems);
#endif
