/*
 * Copyright (c) 2022 Reza Mahdi. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are hereby granted under terms of BSD 3-Clause License.
 * See LICENSE for more information.
 */

#ifndef DNSPIDER_QUEUE_H_
#define DNSPIDER_QUEUE_H_

#if defined(HAVE_CONFIG_H)
#include <config.h>
#endif

#if defined(HAVE_MIMALLOC)
#include <mimalloc.h>
#endif

#include <stdint.h>

typedef struct
{
  volatile uint32_t prod_head; /**< producer's head */
  volatile uint32_t prod_tail; /**< producer's tail */

  /* The queue is memory-bounded. Instead of saving the
   * size of the queue we save the bit mask: assuming
   * a size power of 2 N, we can compute X % N as
   * X & mask for any integer. (where & is faster than %).
   */
  uint32_t prod_mask;

  /* Pad between producer and consumer attributes. This
   * avoids the "false sharing" problem: when we modify
   * and attribute, the whole L1/L2 cache line needs to be
   * updated in all the cores. If at the same time other
   * thread is accessing to the other attributes a conflict
   * will arise. The CPU will know how to fix it but it is
   * going to have a penalty.
   *
   * XXX assuming that the L1 and L2 cache lines are of 64 bytes
   */
  uint32_t _pad1[13];

  volatile uint32_t cons_head; /**< consumer's head */
  volatile uint32_t cons_tail; /**< consumer's tail */
  /* Why again the mask? Having two copies of the mask, each next
   * to the respective head/tail ensures that the head, the tail
   * and the mask of the producer will be in its own L2 cache line
   * avoiding "false sharings"
   */
  uint32_t cons_mask;

  uint32_t _pad2[13];

  uint8_t *data;    /**< beggining of data */
  uint32_t elem_sz; /**< size of an element */
} queue_t;

/************* Single Producer, Single Consumer ******************************/
uint32_t queue_spsc_push (queue_t *q, void *data, uint32_t len, int flags,
                          uint32_t *free_entries_remain);
uint32_t queue_spsc_pop (queue_t *q, void *data, uint32_t len, int flags,
                         uint32_t *ready_entries_remain);

int queue_spsc_init (queue_t *q, uint32_t sz, uint32_t elem_sz);

#if defined(HAVE_MIMALLOC)
int queue_spsc_init_heap (mi_heap_t *heap, queue_t *q, uint32_t sz,
                          uint32_t elem_sz);
#endif

void queue_spsc_destroy (queue_t *q);

uint32_t queue_spsc_ready (queue_t *q);
uint32_t queue_spsc_free (queue_t *q);

#endif
