/*
 * Copyright (c) 2022 Reza Mahdi. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are hereby granted under terms of BSD 3-Clause License.
 * See LICENSE for more information.
 */

#ifndef TEST_H
#define TEST_H

#ifdef __GNUC__
#define TEST_IMPL_FORMAT_PRINTF(a, b) __attribute__ ((format (printf, a, b)))
#else
#define TEST_IMPL_FORMAT_PRINTF(a, b)
#endif

#include <inttypes.h> /* intmax_t, uintmax_t, PRI* */
#include <stddef.h>   /* size_t */

typedef void (*test_nullary_run_func) (void);
typedef void (*test_unary_run_func) (void *);
typedef void (*test_setup_func) (void *);
typedef void (*test_teardown_func) (void *);

union test_run_func_union
{
  test_nullary_run_func nullary;
  test_unary_run_func unary;
};

#define TEST_IMPL_PRAGMA(x) _Pragma (#x)

#if defined(__GNUC__)
#if defined(__clang__) || __GNUC__ > 4                                        \
    || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)
/* the GCC argument will work for both gcc and clang  */
#define TEST_IMPL_DIAG_PUSH_IGNORED(w)                                       \
  TEST_IMPL_PRAGMA (GCC diagnostic push)                                     \
  TEST_IMPL_PRAGMA (GCC diagnostic ignored "-W" #w)
#define TEST_IMPL_DIAG_POP() TEST_IMPL_PRAGMA (GCC diagnostic pop)
#else
/* the push/pop functionality wasn't in gcc until 4.6, fallback to "ignored" */
#define TEST_IMPL_DIAG_PUSH_IGNORED(w)                                       \
  TEST_IMPL_PRAGMA (GCC diagnostic ignored "-W" #w)
#define TEST_IMPL_DIAG_POP()
#endif
#else
/* leave them out entirely for non-GNUC compilers  */
#define TEST_IMPL_DIAG_PUSH_IGNORED(w)
#define TEST_IMPL_DIAG_POP()
#endif

struct test
{
  const char *ssname; // suite name
  const char *ttname; // test name
  union test_run_func_union run;

  void *data;
  test_setup_func *setup;
  test_teardown_func *teardown;

  int skip;

  unsigned int magic;
};

#define TEST_IMPL_NAME(name) test_##name
#define TEST_IMPL_FNAME(sname, tname) TEST_IMPL_NAME (sname##_##tname##_run)
#define TEST_IMPL_TNAME(sname, tname) TEST_IMPL_NAME (sname##_##tname)
#define TEST_IMPL_DATA_SNAME(sname) TEST_IMPL_NAME (sname##_data)
#define TEST_IMPL_DATA_TNAME(sname, tname)                                   \
  TEST_IMPL_NAME (sname##_##tname##_data)
#define TEST_IMPL_SETUP_FNAME(sname) TEST_IMPL_NAME (sname##_setup)
#define TEST_IMPL_SETUP_FPNAME(sname) TEST_IMPL_NAME (sname##_setup_ptr)
#define TEST_IMPL_SETUP_TPNAME(sname, tname)                                 \
  TEST_IMPL_NAME (sname##_##tname##_setup_ptr)
#define TEST_IMPL_TEARDOWN_FNAME(sname) TEST_IMPL_NAME (sname##_teardown)
#define TEST_IMPL_TEARDOWN_FPNAME(sname)                                     \
  TEST_IMPL_NAME (sname##_teardown_ptr)
#define TEST_IMPL_TEARDOWN_TPNAME(sname, tname)                              \
  TEST_IMPL_NAME (sname##_##tname##_teardown_ptr)

#define TEST_IMPL_MAGIC (0xdeadbeef)
#ifdef __APPLE__
#define TEST_IMPL_SECTION                                                    \
  __attribute__ ((used, section ("__DATA, .test"), aligned (1)))
#else
#define TEST_IMPL_SECTION                                                    \
  __attribute__ ((used, section (".test"), aligned (1)))
#endif

#define TEST_IMPL_STRUCT(sname, tname, tskip, tdata, tsetup, tteardown)      \
  static struct test TEST_IMPL_TNAME (sname, tname) TEST_IMPL_SECTION = {  \
    #sname,                                                                   \
    #tname,                                                                   \
    { (test_nullary_run_func)TEST_IMPL_FNAME (sname, tname) },              \
    tdata,                                                                    \
    (test_setup_func *)tsetup,                                               \
    (test_teardown_func *)tteardown,                                         \
    tskip,                                                                    \
    TEST_IMPL_MAGIC,                                                         \
  }

#define TEST_SETUP(sname)                                                    \
  static void TEST_IMPL_SETUP_FNAME (sname) (                                \
      struct TEST_IMPL_DATA_SNAME (sname) * data);                           \
  static void (*TEST_IMPL_SETUP_FPNAME (sname)) (                            \
      struct TEST_IMPL_DATA_SNAME (sname) *)                                 \
      = &TEST_IMPL_SETUP_FNAME (sname);                                      \
  static void TEST_IMPL_SETUP_FNAME (sname) (                                \
      struct TEST_IMPL_DATA_SNAME (sname) * data)

#define TEST_TEARDOWN(sname)                                                 \
  static void TEST_IMPL_TEARDOWN_FNAME (sname) (                             \
      struct TEST_IMPL_DATA_SNAME (sname) * data);                           \
  static void (*TEST_IMPL_TEARDOWN_FPNAME (sname)) (                         \
      struct TEST_IMPL_DATA_SNAME (sname) *)                                 \
      = &TEST_IMPL_TEARDOWN_FNAME (sname);                                   \
  static void TEST_IMPL_TEARDOWN_FNAME (sname) (                             \
      struct TEST_IMPL_DATA_SNAME (sname) * data)

#define TEST_DATA(sname)                                                     \
  struct TEST_IMPL_DATA_SNAME (sname);                                       \
  static void (*TEST_IMPL_SETUP_FPNAME (sname)) (                            \
      struct TEST_IMPL_DATA_SNAME (sname) *);                                \
  static void (*TEST_IMPL_TEARDOWN_FPNAME (sname)) (                         \
      struct TEST_IMPL_DATA_SNAME (sname) *);                                \
  struct TEST_IMPL_DATA_SNAME (sname)

#define TEST_IMPL_TEST(sname, tname, tskip)                                 \
  static void TEST_IMPL_FNAME (sname, tname) (void);                         \
  TEST_IMPL_STRUCT (sname, tname, tskip, NULL, NULL, NULL);                  \
  static void TEST_IMPL_FNAME (sname, tname) (void)

#define TEST_IMPL_TEST2(sname, tname, tskip)                                \
  static struct TEST_IMPL_DATA_SNAME (sname)                                 \
      TEST_IMPL_DATA_TNAME (sname, tname);                                   \
  static void TEST_IMPL_FNAME (sname, tname) (                               \
      struct TEST_IMPL_DATA_SNAME (sname) * data);                           \
  TEST_IMPL_STRUCT (                                                         \
      sname, tname, tskip, &TEST_IMPL_DATA_TNAME (sname, tname),             \
      &TEST_IMPL_SETUP_FPNAME (sname), &TEST_IMPL_TEARDOWN_FPNAME (sname)); \
  static void TEST_IMPL_FNAME (sname, tname) (                               \
      struct TEST_IMPL_DATA_SNAME (sname) * data)

void TEST_LOG (const char *fmt, ...) TEST_IMPL_FORMAT_PRINTF (1, 2);
void TEST_ERR (const char *fmt, ...)
    TEST_IMPL_FORMAT_PRINTF (1, 2); // doesn't return

#define TEST(sname, tname) TEST_IMPL_TEST (sname, tname, 0)
#define TEST_SKIP(sname, tname) TEST_IMPL_TEST (sname, tname, 1)

#define TEST2(sname, tname) TEST_IMPL_TEST2 (sname, tname, 0)
#define TEST2_SKIP(sname, tname) TEST_IMPL_TEST2 (sname, tname, 1)

void assert_str (const char *exp, const char *real, const char *caller,
                 int line);
#define ASSERT_STR(exp, real) assert_str (exp, real, __FILE__, __LINE__)

void assert_wstr (const wchar_t *exp, const wchar_t *real, const char *caller,
                  int line);
#define ASSERT_WSTR(exp, real) assert_wstr (exp, real, __FILE__, __LINE__)

void assert_data (const unsigned char *exp, size_t expsize,
                  const unsigned char *real, size_t realsize,
                  const char *caller, int line);
#define ASSERT_DATA(exp, expsize, real, realsize)                             \
  assert_data (exp, expsize, real, realsize, __FILE__, __LINE__)

void assert_equal (intmax_t exp, intmax_t real, const char *caller, int line);
#define ASSERT_EQUAL(exp, real) assert_equal (exp, real, __FILE__, __LINE__)

void assert_equal_u (uintmax_t exp, uintmax_t real, const char *caller,
                     int line);
#define ASSERT_EQUAL_U(exp, real)                                             \
  assert_equal_u (exp, real, __FILE__, __LINE__)

#define ASSERT_EQUAL_PTR(exp, real)                                           \
  assert_equal_u((uintmax_t)exp, (uintmax_t) real, __FILE__, __LINE__)

void assert_not_equal (intmax_t exp, intmax_t real, const char *caller,
                       int line);
#define ASSERT_NOT_EQUAL(exp, real)                                           \
  assert_not_equal (exp, real, __FILE__, __LINE__)

void assert_not_equal_u (uintmax_t exp, uintmax_t real, const char *caller,
                         int line);
#define ASSERT_NOT_EQUAL_U(exp, real)                                         \
  assert_not_equal_u (exp, real, __FILE__, __LINE__)

void assert_interval (intmax_t exp1, intmax_t exp2, intmax_t real,
                      const char *caller, int line);
#define ASSERT_INTERVAL(exp1, exp2, real)                                     \
  assert_interval (exp1, exp2, real, __FILE__, __LINE__)

void assert_null (void *real, const char *caller, int line);
#define ASSERT_NULL(real) assert_null ((void *)real, __FILE__, __LINE__)

void assert_not_null (const void *real, const char *caller, int line);
#define ASSERT_NOT_NULL(real) assert_not_null (real, __FILE__, __LINE__)

void assert_true (int real, const char *caller, int line);
#define ASSERT_TRUE(real) assert_true (real, __FILE__, __LINE__)

void assert_false (int real, const char *caller, int line);
#define ASSERT_FALSE(real) assert_false (real, __FILE__, __LINE__)

void assert_fail (const char *caller, int line);
#define ASSERT_FAIL() assert_fail (__FILE__, __LINE__)

void assert_dbl_near (double exp, double real, double tol, const char *caller,
                      int line);
#define ASSERT_DBL_NEAR(exp, real)                                            \
  assert_dbl_near (exp, real, 1e-4, __FILE__, __LINE__)
#define ASSERT_DBL_NEAR_TOL(exp, real, tol)                                   \
  assert_dbl_near (exp, real, tol, __FILE__, __LINE__)

void assert_dbl_far (double exp, double real, double tol, const char *caller,
                     int line);
#define ASSERT_DBL_FAR(exp, real)                                             \
  assert_dbl_far (exp, real, 1e-4, __FILE__, __LINE__)
#define ASSERT_DBL_FAR_TOL(exp, real, tol)                                    \
  assert_dbl_far (exp, real, tol, __FILE__, __LINE__)

#ifdef TEST_MAIN

#include <setjmp.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <wchar.h>

static size_t test_errorsize;
static char *test_errormsg;
#define MSG_SIZE 4096
static char test_errorbuffer[MSG_SIZE];
static jmp_buf test_err;
static int color_output = 1;
static const char *suite_name;

typedef int (*test_filter_func) (struct test *);

#define ANSI_BLACK "\033[0;30m"
#define ANSI_RED "\033[0;31m"
#define ANSI_GREEN "\033[0;32m"
#define ANSI_YELLOW "\033[0;33m"
#define ANSI_BLUE "\033[0;34m"
#define ANSI_MAGENTA "\033[0;35m"
#define ANSI_CYAN "\033[0;36m"
#define ANSI_GREY "\033[0;37m"
#define ANSI_DARKGREY "\033[01;30m"
#define ANSI_BRED "\033[01;31m"
#define ANSI_BGREEN "\033[01;32m"
#define ANSI_BYELLOW "\033[01;33m"
#define ANSI_BBLUE "\033[01;34m"
#define ANSI_BMAGENTA "\033[01;35m"
#define ANSI_BCYAN "\033[01;36m"
#define ANSI_WHITE "\033[01;37m"
#define ANSI_NORMAL "\033[0m"

TEST (suite, test) {}

static void vprint_errormsg (const char *const fmt, va_list ap)
    TEST_IMPL_FORMAT_PRINTF (1, 0);
static void print_errormsg (const char *const fmt, ...)
    TEST_IMPL_FORMAT_PRINTF (1, 2);

static void
vprint_errormsg (const char *const fmt, va_list ap)
{
  // (v)snprintf returns the number that would have been written
  const int ret = vsnprintf (test_errormsg, test_errorsize, fmt, ap);
  if (ret < 0)
    {
      test_errormsg[0] = 0x00;
    }
  else
    {
      const size_t size = (size_t)ret;
      const size_t s
          = (test_errorsize <= size ? size - test_errorsize : size);
      // test_errorsize may overflow at this point
      test_errorsize -= s;
      test_errormsg += s;
    }
}

static void
print_errormsg (const char *const fmt, ...)
{
  va_list argp;
  va_start (argp, fmt);
  vprint_errormsg (fmt, argp);
  va_end (argp);
}

static void
msg_start (const char *color, const char *title)
{
  if (color_output)
    {
      print_errormsg ("%s", color);
    }
  print_errormsg ("  %s: ", title);
}

static void
msg_end (void)
{
  if (color_output)
    {
      print_errormsg (ANSI_NORMAL);
    }
  print_errormsg ("\n");
}

void
TEST_LOG (const char *fmt, ...)
{
  va_list argp;
  msg_start (ANSI_BLUE, "LOG");

  va_start (argp, fmt);
  vprint_errormsg (fmt, argp);
  va_end (argp);

  msg_end ();
}

TEST_IMPL_DIAG_PUSH_IGNORED (missing - noreturn)

void
TEST_ERR (const char *fmt, ...)
{
  va_list argp;
  msg_start (ANSI_YELLOW, "ERR");

  va_start (argp, fmt);
  vprint_errormsg (fmt, argp);
  va_end (argp);

  msg_end ();
  longjmp (test_err, 1);
}

TEST_IMPL_DIAG_POP ()

void
assert_str (const char *exp, const char *real, const char *caller, int line)
{
  if ((exp == NULL && real != NULL) || (exp != NULL && real == NULL)
      || (exp && real && strcmp (exp, real) != 0))
    {
      TEST_ERR ("%s:%d  expected '%s', got '%s'", caller, line, exp, real);
    }
}

void
assert_wstr (const wchar_t *exp, const wchar_t *real, const char *caller,
             int line)
{
  if ((exp == NULL && real != NULL) || (exp != NULL && real == NULL)
      || (exp && real && wcscmp (exp, real) != 0))
    {
      TEST_ERR ("%s:%d  expected '%ls', got '%ls'", caller, line, exp, real);
    }
}

void
assert_data (const unsigned char *exp, size_t expsize,
             const unsigned char *real, size_t realsize, const char *caller,
             int line)
{
  size_t i;
  if (expsize != realsize)
    {
      TEST_ERR ("%s:%d  expected %" PRIuMAX " bytes, got %" PRIuMAX, caller,
                 line, (uintmax_t)expsize, (uintmax_t)realsize);
    }
  for (i = 0; i < expsize; i++)
    {
      if (exp[i] != real[i])
        {
          TEST_ERR ("%s:%d expected 0x%02x at offset %" PRIuMAX " got 0x%02x",
                     caller, line, exp[i], (uintmax_t)i, real[i]);
        }
    }
}

void
assert_equal (intmax_t exp, intmax_t real, const char *caller, int line)
{
  if (exp != real)
    {
      TEST_ERR ("%s:%d  expected %" PRIdMAX ", got %" PRIdMAX, caller, line,
                 exp, real);
    }
}

void
assert_equal_u (uintmax_t exp, uintmax_t real, const char *caller, int line)
{
  if (exp != real)
    {
      TEST_ERR ("%s:%d  expected %" PRIuMAX ", got %" PRIuMAX, caller, line,
                 exp, real);
    }
}

void
assert_not_equal (intmax_t exp, intmax_t real, const char *caller, int line)
{
  if ((exp) == (real))
    {
      TEST_ERR ("%s:%d  should not be %" PRIdMAX, caller, line, real);
    }
}

void
assert_not_equal_u (uintmax_t exp, uintmax_t real, const char *caller,
                    int line)
{
  if ((exp) == (real))
    {
      TEST_ERR ("%s:%d  should not be %" PRIuMAX, caller, line, real);
    }
}

void
assert_interval (intmax_t exp1, intmax_t exp2, intmax_t real,
                 const char *caller, int line)
{
  if (real < exp1 || real > exp2)
    {
      TEST_ERR ("%s:%d  expected %" PRIdMAX "-%" PRIdMAX ", got %" PRIdMAX,
                 caller, line, exp1, exp2, real);
    }
}

void
assert_dbl_near (double exp, double real, double tol, const char *caller,
                 int line)
{
  double diff = exp - real;
  double absdiff = diff;
  /* avoid using fabs and linking with a math lib */
  if (diff < 0)
    {
      absdiff *= -1;
    }
  if (absdiff > tol)
    {
      TEST_ERR ("%s:%d  expected %0.3e, got %0.3e (diff %0.3e, tol %0.3e)",
                 caller, line, exp, real, diff, tol);
    }
}

void
assert_dbl_far (double exp, double real, double tol, const char *caller,
                int line)
{
  double diff = exp - real;
  double absdiff = diff;
  /* avoid using fabs and linking with a math lib */
  if (diff < 0)
    {
      absdiff *= -1;
    }
  if (absdiff <= tol)
    {
      TEST_ERR ("%s:%d  expected %0.3e, got %0.3e (diff %0.3e, tol %0.3e)",
                 caller, line, exp, real, diff, tol);
    }
}

void
assert_null (void *real, const char *caller, int line)
{
  if ((real) != NULL)
    {
      TEST_ERR ("%s:%d  should be NULL", caller, line);
    }
}

void
assert_not_null (const void *real, const char *caller, int line)
{
  if (real == NULL)
    {
      TEST_ERR ("%s:%d  should not be NULL", caller, line);
    }
}

void
assert_true (int real, const char *caller, int line)
{
  if ((real) == 0)
    {
      TEST_ERR ("%s:%d  should be true", caller, line);
    }
}

void
assert_false (int real, const char *caller, int line)
{
  if ((real) != 0)
    {
      TEST_ERR ("%s:%d  should be false", caller, line);
    }
}

void
assert_fail (const char *caller, int line)
{
  TEST_ERR ("%s:%d  shouldn't come here", caller, line);
}

static int
suite_all (struct test *t)
{
  (void)t; // fix unused parameter warning
  return 1;
}

static int
suite_filter (struct test *t)
{
  return strncmp (suite_name, t->ssname, strlen (suite_name)) == 0;
}

static uint64_t
getCurrentTime (void)
{
  struct timeval now;
  gettimeofday (&now, NULL);
  uint64_t now64 = (uint64_t)now.tv_sec;
  now64 *= 1000000;
  now64 += ((uint64_t)now.tv_usec);
  return now64;
}

static void
color_print (const char *color, const char *text)
{
  if (color_output)
    printf ("%s%s" ANSI_NORMAL "\n", color, text);
  else
    printf ("%s\n", text);
}

#ifdef TEST_SEGFAULT
#include <signal.h>
static void
sighandler (int signum)
{
  const char msg_color[]
      = ANSI_BRED "[SIGSEGV: Segmentation fault]" ANSI_NORMAL "\n";
  const char msg_nocolor[] = "[SIGSEGV: Segmentation fault]\n";

  const char *msg = color_output ? msg_color : msg_nocolor;
  write (STDOUT_FILENO, msg, strlen (msg));

  /* "Unregister" the signal handler and send the signal back to the process
   * so it can terminate as expected */
  signal (signum, SIG_DFL);
  kill (getpid (), signum);
}
#endif

int test_main (int argc, const char *argv[]);

__attribute__ ((no_sanitize_address)) int
test_main (int argc, const char *argv[])
{
  static int total = 0;
  static int num_ok = 0;
  static int num_fail = 0;
  static int num_skip = 0;
  static int idx = 1;
  static test_filter_func filter = suite_all;

#ifdef TEST_SEGFAULT
  signal (SIGSEGV, sighandler);
#endif

  if (argc == 2)
    {
      suite_name = argv[1];
      filter = suite_filter;
    }
#ifdef TEST_NO_COLORS
  color_output = 0;
#else
  color_output = isatty (1);
#endif
  uint64_t t1 = getCurrentTime ();

  struct test *test_begin = &TEST_IMPL_TNAME (suite, test);
  struct test *test_end = &TEST_IMPL_TNAME (suite, test);
  // find begin and end of section by comparing magics
  while (1)
    {
      struct test *t = test_begin - 1;
      if (t->magic != TEST_IMPL_MAGIC)
        break;
      test_begin--;
    }
  while (1)
    {
      struct test *t = test_end + 1;
      if (t->magic != TEST_IMPL_MAGIC)
        break;
      test_end++;
    }
  test_end++; // end after last one

  static struct test *test;
  for (test = test_begin; test != test_end; test++)
    {
      if (test == &TEST_IMPL_TNAME (suite, test))
        continue;
      if (filter (test))
        total++;
    }

  for (test = test_begin; test != test_end; test++)
    {
      if (test == &TEST_IMPL_TNAME (suite, test))
        continue;
      if (filter (test))
        {
          test_errorbuffer[0] = 0;
          test_errorsize = MSG_SIZE - 1;
          test_errormsg = test_errorbuffer;
          printf ("TEST %d/%d %s:%s ", idx, total, test->ssname, test->ttname);
          fflush (stdout);
          if (test->skip)
            {
              color_print (ANSI_BYELLOW, "[SKIPPED]");
              num_skip++;
            }
          else
            {
              int result = setjmp (test_err);
              if (result == 0)
                {
                  if (test->setup && *test->setup)
                    (*test->setup) (test->data);
                  if (test->data)
                    test->run.unary (test->data);
                  else
                    test->run.nullary ();
                  if (test->teardown && *test->teardown)
                    (*test->teardown) (test->data);
                    // if we got here it's ok
#ifdef TEST_COLOR_OK
                  color_print (ANSI_BGREEN, "[OK]");
#else
                  printf ("[OK]\n");
#endif
                  num_ok++;
                }
              else
                {
                  color_print (ANSI_BRED, "[FAIL]");
                  num_fail++;
                }
              if (test_errorsize != MSG_SIZE - 1)
                printf ("%s", test_errorbuffer);
            }
          idx++;
        }
    }
  uint64_t t2 = getCurrentTime ();

  const char *color = (num_fail) ? ANSI_BRED : ANSI_GREEN;
  char results[80];
  snprintf (results, sizeof (results),
            "RESULTS: %d tests (%d ok, %d failed, %d skipped) ran in %" PRIu64
            " ms",
            total, num_ok, num_fail, num_skip, (t2 - t1) / 1000);
  color_print (color, results);
  return num_fail;
}

#endif

#endif
